import copy
from typing import Dict

from . import CharacterRenderer, BasicCharacterRenderer, UnderlineCharacterRenderer, ShadowCharacterRenderer, OutlineCharacterRenderer
from gui.text import TextStyle


class CharacterRendererFactory:
    def __init__(self):
        self.__characterRenderers = {}  # type: Dict[TextStyle, CharacterRenderer]

    @classmethod
    def getInstance(cls):
        if not hasattr(cls, "instance"):
            cls.instance = CharacterRendererFactory()
        return cls.instance

    def create(self, style: TextStyle) -> CharacterRenderer:
        if style in self.__characterRenderers:
            return self.__characterRenderers[style]

        # All we can create using Pygame font
        currentStyle = TextStyle(style.fontName, style.tileHeight, style.color)
        currentStyle.italic = style.italic
        currentStyle.bold = style.bold
        if currentStyle in self.__characterRenderers:
            characterRenderer = self.__characterRenderers[currentStyle]
        else:
            characterRenderer = BasicCharacterRenderer(currentStyle)
            self.__characterRenderers[currentStyle] = characterRenderer

        # Add underline
        if style.flags & TextStyle.UNDERLINE:
            currentStyle.underline = True
            if currentStyle in self.__characterRenderers:
                characterRenderer = self.__characterRenderers[currentStyle]
            else:
                characterRenderer = UnderlineCharacterRenderer(currentStyle, characterRenderer)
                self.__characterRenderers[currentStyle] = characterRenderer

        # Outline
        if style.flags & TextStyle.OUTLINE:
            currentStyle.outline = True
            currentStyle.outlineColor = style.outlineColor
            currentStyle.outlineThickness = style.outlineThickness
            if currentStyle in self.__characterRenderers:
                characterRenderer = self.__characterRenderers[currentStyle]
            else:
                characterRenderer = OutlineCharacterRenderer(currentStyle, characterRenderer)
                self.__characterRenderers[currentStyle] = characterRenderer

        # Shadow
        if style.flags & TextStyle.SHADOW:
            currentStyle.shadow = True
            currentStyle.shadowColor = style.shadowColor
            currentStyle.shadowShift = style.shadowShift
            if currentStyle in self.__characterRenderers:
                characterRenderer = self.__characterRenderers[currentStyle]
            else:
                # We need a renderer with the shadow color and no shadow
                shadowColorStyle = copy.deepcopy(currentStyle)
                shadowColorStyle.shadow = False
                shadowColorStyle.color = style.shadowColor
                shadowColorStyle.outlineColor = style.shadowColor
                shadowCharacterRenderer = self.create(shadowColorStyle)
                # The text + shadow renderer
                characterRenderer = ShadowCharacterRenderer(currentStyle, characterRenderer, shadowCharacterRenderer)
                self.__characterRenderers[currentStyle] = characterRenderer

        return characterRenderer
