import os
from typing import List

import numpy as np
from tmx import TileMap, Layer, LayerTile, Tileset


class LevelLoader:

    def __init__(self, fileName: str):
        self.__fileName = fileName

        # Load map
        if not os.path.exists(fileName):
            raise RuntimeError("No file {}".format(fileName))
        self.__tileMap = TileMap.load(fileName)

        # Check main properties
        if self.__tileMap.orientation != "orthogonal":
            raise RuntimeError("Error in {}: invalid orientation".format(fileName))

    @property
    def fileName(self) -> str:
        return self.__fileName

    @property
    def width(self) -> int:
        return self.__tileMap.width

    @property
    def height(self) -> int:
        return self.__tileMap.height

    def decodeLayer(self, layerIndex) -> (np.ndarray, str, int, int):
        # Layer
        if layerIndex >= len(self.__tileMap.layers):
            raise RuntimeError("Error in {}: no layer {}".format(self.fileName, layerIndex))
        layer = self.__tileMap.layers[layerIndex]  # type: Layer

        # Tiles
        tiles = layer.tiles  # type: List[LayerTile]
        if len(tiles) != self.width * self.height:
            raise RuntimeError("Error in {}: invalid tiles count".format(self.fileName))

        # Tilesets
        tilesets = self.__tileMap.tilesets  # type: List[Tileset]
        if len(tilesets) == 0:
            raise RuntimeError("Error in {}: no tilesets".format(self.fileName))

        # Guess which tileset is used by this layer
        gid = None
        for tile in tiles:
            if tile.gid != 0:
                gid = tile.gid
                break
        if gid is None:
            tileset = tilesets[0]
        else:
            tileset = None
            for t in tilesets:
                if t.firstgid <= gid < t.firstgid + t.tilecount:
                    tileset = t
                    break
            if tileset is None:
                raise RuntimeError("Error in {}: no corresponding tileset".format(self.fileName))

        # Check the associated tileset
        if tileset.columns <= 0:
            raise RuntimeError("Error in {}: invalid columns count".format(self.fileName))
        if tileset.image.data is not None:
            raise RuntimeError("Error in {}: embedded tileset image is not supported".format(self.fileName))

        # Build a array of  for the associated tileset
        tileCoords = np.zeros((self.width, self.height, 2), dtype=np.int32)
        for y in range(self.height):
            for x in range(self.width):
                tile = layer.tiles[x + y * self.width]
                if tile.gid == 0:
                    tileCoords[x, y, 0] = tileset.columns - 1
                    tileCoords[x, y, 1] = tileset.columns - 1
                else:
                    lid = tile.gid - tileset.firstgid
                    if lid < 0 or lid >= tileset.tilecount:
                        raise RuntimeError("Error in {}: invalid tile id".format(self.fileName))
                    tileCoords[x, y, 0] = lid % tileset.columns
                    tileCoords[x, y, 1] = lid // tileset.columns

        return tileCoords, tileset.image.source, tileset.tilewidth, tileset.tileheight
